// Fill out your copyright notice in the Description page of Project Settings.


#include "AdventureMap.h"
#include "HexTile.h"
#include "AdventurePlayerController.h"
#include "Kismet/GameplayStatics.h"
#include "Algo/Reverse.h"

// Sets default values
AAdventureMap::AAdventureMap()
{
	FHexVector NBs[] = { NNE, E, SSE, SSW, W, NNW };
	NeighborUnitVectors.Append(NBs, UE_ARRAY_COUNT(NBs));
	FHexVector DNBs[] = { N, ENE, ESE, S, WSW, WNW };
	DiagonalUnitVectors.Append(DNBs, UE_ARRAY_COUNT(DNBs));
}

// Called when the game starts or when spawned
void AAdventureMap::BeginPlay()
{
	Super::BeginPlay();
	World = GetWorld();

	if (IsValid(BaseTileClass)) { 
		MakeGrid();
	}
}

// Called once on Begin Play
void AAdventureMap::MakeGrid() 
{	
	FVector NextHexAt = FVector();
	float HexWidth = sqrt(3) * TileSize;
	int QOffset = 0;

	for (int r = 1; r <= GridSize; r++) {
		float XOffset = 0.f;

		if (r % 2 != 0)	{ if (r > 1) { QOffset--; } }
		else { XOffset = HexWidth / 2; }

		for (int q = 1; q <= GridSize; q++) {
			NextHexAt.X = XOffset + (HexWidth * q);
			NextHexAt.Y = TileSize * 1.5f * r;
			NextHexAt.Z = 0.f;
			FTransform SpawnTransform = FTransform(NextHexAt);

			AHexTile* Tile = World->SpawnActor<AHexTile>(BaseTileClass, SpawnTransform);
			Grid.Add(Tile);

			Tile->Q = q - 1 + QOffset;
			Tile->R = r - 1;
		}
	}

	for (auto& tile : Grid)	{
		tile->Index = GridIndex(tile->Q, tile->R);
	}

	bHexGridReady = true;
}

// Every Hex Tile's index within the Grid Array can be derived from its Q and R coordinates
int32 AAdventureMap::GridIndex(int32 qAxial, int32 rAxial)
{
	/*
	*	The Q axis is (i.e. columns are) oriented diagonally.
	*	The Hex Grid has a rough square shape, hence the Q coordinates must be offset by -1 every other row.
	*/
	int32 column = qAxial + FMath::FloorToInt(rAxial / 2);
	return (rAxial * GridSize) + column;
}

AHexTile* AAdventureMap::RandomHex()
{
	int32 RandHex = FMath::RandRange(0, GridSize*GridSize-1);
	return Grid[RandHex];
}

/*
*	Add two TArray<TPair<int32, int32>> members containing the Cardinal Directions
	(one for immediate neighbors, one for diagonals)
	{ fill them in AAdventureMap::AAdventureMap }

*	This function instead returns
		TMap<bool bDiag, AHexTile* Neighbor>
*/
TArray<AHexTile*> AAdventureMap::Neighbors(AHexTile* OfHex, bool bFreeOnly = false)
{
	TArray<AHexTile*> Results;
	for (auto& V : NeighborUnitVectors) {
		int32 I = GridIndex(OfHex->Q + V.Q, OfHex->R + V.R);
		if (Grid.IsValidIndex(I)) { 
			AHexTile* R = Grid[I];
			Results.Add(Grid[I]);
			if (bFreeOnly && !R->bFree) { Results.Remove(R); }
		}
	}
	for (auto& R : Results) {
		if (bFreeOnly && !R->bFree) { Results.Remove(R); }
	}
	return Results;
}

TArray<AHexTile*> AAdventureMap::FreeDiags(AHexTile* OfHex)
{
	TArray<AHexTile*> Results;
	for (auto& V : DiagonalUnitVectors) {
		int32 I = GridIndex(OfHex->Q + V.Q, OfHex->R + V.R);
		if (!Grid.IsValidIndex(I)) { continue; }
		// if (!bFreeOnly) { if (Grid[I]->Distance(OfHex) == 1) { Results.Add(Grid[I]); } }
		else {
			bool bReachable = true;
			for (auto& PotentialBlock : Neighbors(OfHex)) {
				if (PotentialBlock->Distance(Grid[I]) != 1) { continue; }
				if (!PotentialBlock->bFree) {
					bReachable = false;
					break;
				}
			}
			if (bReachable) {
				Results.Add(Grid[I]);
			}
		}
	}
	return Results;
}

TSet<AHexTile*> AAdventureMap::BreadthFirstSearch(AHexTile* Start, int32 Radius)
{
	TSet<AHexTile*> Results;
	TArray<AHexTile*> ToExamine;
	TSet<AHexTile*> Processed;
	Results.Add(Start);
	ToExamine.Add(Start);

	while (!ToExamine.IsEmpty()) {
		AHexTile* Candidate = ToExamine[0];
		Processed.Add(Candidate);
		ToExamine.Remove(Candidate);

		for (AHexTile* Neighbor : Neighbors(Candidate)) {
			if (Neighbor->Distance(Candidate) > 1)	{ continue; }
			if (Processed.Contains(Neighbor))		{ continue; }
			if (Neighbor->Distance(Start) > Radius)	{ continue; }

			ToExamine.Add(Neighbor);
			Results.Add(Neighbor);
		}
	}
	return Results;
}

TArray<AHexTile*> AAdventureMap::FindPathAStar(AHexTile* Start, AHexTile* Goal, bool bDiags)
{
	TArray<AHexTile*> ToExamine;
	TSet<AHexTile*> Processed;
	ToExamine.Add(Start);

	while (!ToExamine.IsEmpty()) {
		AHexTile* Candidate = ToExamine[0];
		ToExamine.Remove(Candidate);
		// try for Hex with lower estimatet (F)cost
		for (auto& t : ToExamine) {
			t->FCost = t->GCost + t->HCost;
			if (t->FCost < Candidate->FCost || t->FCost == Candidate->FCost && t->HCost < Candidate->HCost) { Candidate = t; }
		}

		Processed.Add(Candidate);
		// exit
		if (Candidate == Goal) { break; }

		// expand frontier & adjust path data
		for (AHexTile* Neighbor : Neighbors(Candidate, true)) {
			if (Neighbor->Distance(Candidate) > 1) { continue; }
			if (Processed.Contains(Neighbor)) { continue; }

			bool bInToExamine = ToExamine.Contains(Neighbor);
			float NewGCost = Candidate->GCost + Neighbor->MoveCost * 10.f;

			if (NewGCost < Neighbor->GCost || !bInToExamine) {
				Neighbor->GCost = NewGCost;
				Neighbor->CameFrom = Candidate; // chain

				if (!bInToExamine) {
					Neighbor->HCost = Neighbor->Distance(Goal) * 10.f;
					ToExamine.Add(Neighbor);
				}
			}
		}
		if (bDiags) { // right now the heuristic for HCost (Distance func) does NOT take diagonals into account
			for (AHexTile* Diag : FreeDiags(Candidate)) {
				if (Diag->Distance(Candidate) > 2) { continue; }
				if (!Diag->bFree) { continue; }
				if (Processed.Contains(Diag)) { continue; }

				bool bInToExamine = ToExamine.Contains(Diag);
				float NewGCost = Candidate->GCost + Diag->MoveCost * 10.f;

				if (NewGCost < Diag->GCost || !bInToExamine) {
					Diag->GCost = NewGCost;
					Diag->CameFrom = Candidate; // chain

					if (!bInToExamine) {
						Diag->HCost = Diag->Distance(Goal) * 10.f;
						ToExamine.Add(Diag);
					}
				}
			}
		}
	}
	TArray<AHexTile*> Path;
	if (!IsValid(Goal->CameFrom)) { return Path; }
	AHexTile* iPathNode = Goal;
	while (iPathNode != Start) {
		Path.Emplace(iPathNode);
		iPathNode = iPathNode->CameFrom;
	}
	Algo::Reverse(Path);
	return Path;
}